package io.core.messaging; 

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Vector;
import salsa.language.*;
import wwc.messaging.Theater;
import io.core.decision.DecisionAgent;
import io.core.messaging.*;
import io.core.server.*;

public abstract class PeerTheater extends Theater implements IOTheater {
	static private Hashtable messageQueue = new Hashtable();
	static private Vector neighbors = new Vector();
	static private Vector peerServerList = new Vector();
	static private TheaterLocator location;

	static public DecisionAgent decider;
	static public ProtocolHandler protocolHandler;

	static private int maxConnections = 10;
	static String localhostName = null;
	static public int port;


	public void startTheater() {
		System.out.println("Theater listening on: " + location.toString());
		super.startTheater(port, null);
	}

	//initialize takes all the arguments, gets the theaterlocator for
	//this location, and starts a peer server handler if one is desired
	public static void initialize(String[] args) {
		for (int i = 0; i < args.length; i++) {

			if (args[i].equals("-?") || args[i].equals("-h")) 	PeerTheater.helpMessage();
			else if (args[i].equals("-p") || args[i].equals("-c") || args[i].equals("-f")) {
				try {
					if (args[i].equals("-p")) 		port = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-c"))		maxConnections = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-f"))		getSettingsFromFile(args[i+1]);
				} catch (Exception e) {
					if (args[i].equals("-p")) 		PeerTheater.helpMessage("invalid port number : " + args[i+1]);
					else if (args[i].equals("-c"))		PeerTheater.helpMessage("invalid number of max connections : " + args[i+1]);
				}
				i = i + 1;
			}
			else if (args[i].equals("-s")) {
				i++;
				peerServerList = new Vector();
				while ((i < args.length) && ((args[i]).charAt(0) != '-')) {
					peerServerList.add(args[i]);							
					i++;
				}
				i--;
				if (args[i].equals("-s")) PeerTheater.helpMessage("You must specify at least one PeerServer");

			}
			else if (args[i].equals("-l")) {
				localhostName = args[i+1];
				i ++;
			}
			else PeerTheater.helpMessage("Invalid Arguments");
		}

		try {
			String StringHere;
			if (localhostName == null) {
				InetAddress IPHere = InetAddress.getLocalHost();

				StringHere = "rmsp://" + IPHere.getHostAddress() + ":" + port;
			} else {
				StringHere = "rmsp://" + localhostName + ":" + port;
			}

			location = new TheaterLocator(StringHere);
		} catch (Exception e) {
			System.err.println("Error resolving localhost: " + e);
			System.exit(0);
		}

		//	The PeerConnector will connect to the given servers and look for 
		//	The Decider will decide when to migrate actors to other peer theaters, and when to
		//	accept actors from other peer theaters
		if (peerServerList.size() > 0) {
			ServerHandler serverHandler = new ServerHandler(peerServerList, maxConnections, location);
			serverHandler.start();
		}
	}


	private static void getSettingsFromFile(String filename) {
		FileInputStream file;
		byte input[];

		try {
			file = new FileInputStream(filename);

			input = new byte[file.available()];
			file.read(input);
		} catch (Exception e) {
			System.err.println("File Error: " + e);
			return;
		}

		StringTokenizer st = new StringTokenizer(new String(input), "\n ");
		String args[] = new String[st.countTokens()];

		int i = 0;
		while (st.hasMoreTokens()) {
			args[i] = st.nextToken();
			i++;
		}
	}


	private static void helpMessage(String s) {
		System.out.println(s);
		PeerTheater.printHelpMessage();
		System.exit(1);
	}

	private static void helpMessage() {
		PeerTheater.printHelpMessage();
		System.exit(0);
	}

	private static void printHelpMessage() {
		System.out.println("usage :");
		System.out.println("\tjava ...PeerTheater");
		System.out.println("\tjava ...PeerTheater -h");
		System.out.println("\tjava ...PeerTheater -v");
		System.out.println("\tjava ...PeerTheater -p   portNumber");
		System.out.println("\t                    -s   server1 server2 ... serverN");
		System.out.println("\t                    -c   numConnections");
		System.out.println("\t                    -f   filename");
		System.out.println("\t                    -l   localhostName");

		System.out.println("options :");
		System.out.println("\t-h                             : print this message");
		System.out.println("\t-v                             : print version");
		System.out.println("\t-f filename                    : load settings from file");
		System.out.println("\t                                 file is in the same format as command line arguments");
		System.out.println("\t-p portNumber                  : set the listening port to portNumber");
		System.out.println("\t                                 default port number is 3030");
		System.out.println("\t-l localhostName               : set the localhost to the given");
		System.out.println("\t                               : value (to use myrinet)");
		System.out.println("\t-s server1 server2 ... serverN : gets connections to other");
		System.out.println("\t                                 PeerTheaters, from servers 1 through N");
		System.out.println("\t-c numConnections              : specifies the maximum amount of connections");
		System.out.println("\t                                 kept to PeerTheaters, default is 10");
	}

	/**
		The following functions are used by autonomous and universal actors to access
		various information found in the theater
	*/

	public DecisionAgent getDecisionAgent() {
		return decider;
	}

	public ProtocolHandler getProtocolHandler() {
		return protocolHandler;
	}

	public TheaterLocator getLocation() {
		return location;
	}

	public Hashtable getMessageQueue() {
		return messageQueue;
	}

	public Vector getNeighbors() {
		return neighbors;
	}
}
