/** SALSA/World Wide Computer Project
 *  ObjectHandler recieves incoming objects
 *  and determines if they are actors or messages.
 *  upon recieving actors, it will start them up.
 *  messages get sent to the appropriate actor.
 *
 *  author Travis Desell
 */
package io.irandom.messaging;

import java.net.*;
import java.io.*;
import java.util.Vector;
import java.util.Random;
import salsa.language.*;
import io.core.language.*;
import io.core.protocol.P2PPacket;
import io.core.messaging.ProtocolHandler;
import io.core.messaging.P2PTheater;
import io.core.messaging.TheaterLocator;
import io.core.profiling.DecisionAgent;
import io.random.protocol.Terminated;
import io.irandom.protocol.IRandomSteal;
import io.irandom.profiling.IRandomDecisionAgent;


public class IRandomProtocolHandler implements ProtocolHandler {
	private P2PTheater theater;

	int maxTimeToLive = 10;

	public IRandomProtocolHandler(P2PTheater _theater) {
		theater = _theater;
	}

	public void process(P2PPacket recieved) {
		IRandomDecisionAgent decider = (IRandomDecisionAgent)theater.getDecisionAgent();

		if (recieved instanceof IRandomSteal) {
			IRandomSteal iRandomSteal = (IRandomSteal)recieved;

			IOActor candidate = decider.getBestCandidate(iRandomSteal.getOfferedData(), iRandomSteal.getOrigin());

			if ( candidate != null ) {
				//if we should migrate an actor to the origin of this RandomSteal packet,
				//priority send that actor a message to do so.

				System.err.println("Migrating: " + ((UniversalActor)candidate).getUAL().toString() );

				String migrationTarget = iRandomSteal.getOrigin().toString() + ((UniversalActor)candidate).getUAN().getIdentifier();

				Actor[] _targets = { (Actor)candidate };
				String[] _methodNames = { "migrate" };
				Object[][] _arguments = { {migrationTarget} };

				candidate.prioritySend( new Message( (Actor)candidate, _targets, _methodNames, _arguments ) );

				//tell the originating theater that its RandomSteal packet has been terminated
				Terminated terminated = new Terminated(theater.getLocation());
				terminated.send(iRandomSteal.getOrigin());							
			} else {
				if (iRandomSteal.timeToLive > maxTimeToLive) {
					iRandomSteal.timeToLive = maxTimeToLive;
				}
				iRandomSteal.timeToLive--;

				if ((iRandomSteal.timeToLive > 0) && (canSend(iRandomSteal.getOrigin(), theater.getNeighbors()))) {
					//send the random steal packet to a random neighbor that is not the originator
					iRandomSteal.randomSend(theater.getNeighbors());
				} else {
					//tell the originating theater that its RandomSteal packet has been terminated
					Terminated terminated = new Terminated(theater.getLocation());
					terminated.send(iRandomSteal.getOrigin());							
				}
			}
		} else if (recieved instanceof Terminated) {
			decider.stealTerminated();
		}


		//add the originator of this packet to our list of peers,
		//if its not already there
		addPeer( recieved.getOrigin() );
   	}

	private void addPeer(TheaterLocator peer) {
		Vector neighbors = theater.getNeighbors();

		for (int i = 0; i < neighbors.size(); i++) {
			if (peer.equals( (TheaterLocator)neighbors.get(i) )) return;
		}

		theater.getNeighbors().add(peer);
	}

	private boolean canSend(TheaterLocator packetOrigin, Vector targets) {
		if (targets.size() < 1) return false;
		else if ((targets.size() == 1) && (targets.contains(packetOrigin))) return false;
		else return true;
	}
}

