package io.random.decision;

import java.util.Vector;
import salsa.language.ServiceFactory;
import io.core.decision.DecisionAgent;
import io.core.language.*;
import io.core.messaging.*;
import io.random.language.RandomActorState;
import io.random.protocol.RandomSteal;

public class RandomDecisionAgent extends Thread implements DecisionAgent {

	//Waiting determines if the decision agent will originate another random stealing packet.
	private boolean waiting = false;
	private IOTheater theater;

	public RandomDecisionAgent() {
		theater = (IOTheater)ServiceFactory.getTheater();
	}

	public void stealTerminated() {
		waiting = false;
	}

	public void reset() {
		//reset all the autonomous actors in the system
		Vector registry = new Vector(theater.getRegistry().values());

		for (int i = 0; i < registry.size(); i++) {
			Object current = registry.get(i);
			if (current instanceof RandomActorState) {
				((RandomActorState)current).reset();
			}
		}
	}

	public IOActorState getBestCandidate() {
		//returns the best candidate actor to migrate to the given theater

		Object candidate = null;

		synchronized( theater.getRegistry() ) {
			Vector registry = new Vector(theater.getRegistry().values());

			for (int i = 0; i < registry.size(); i++) {
				candidate = registry.get(i);
	
				if (candidate instanceof RandomActorState) {
					if (( !((RandomActorState)candidate).satisfied() ) &&
					    ( !((RandomActorState)candidate).scheduled() )) break;
				}
				candidate = null;
			}
		}

		return (RandomActorState)candidate;
	}

	private void beginSteal() {
		if (!theater.getNeighbors().isEmpty()) {
			System.err.println("Sent a random steal packet");

			waiting = true;

			//create a RandomSteal packet, and send it to a random neighbor
			RandomSteal randomSteal = new RandomSteal(theater.getLocation());
			randomSteal.randomSend(theater.getNeighbors());
		}
	}

	private boolean overloaded() {
		Vector registry = new Vector(theater.getRegistry().values());

		for (int i = 0; i < registry.size(); i++) {
			Object current = registry.get(i);
			if (current instanceof RandomActorState) {
				if ( !((RandomActorState)current).satisfied() ) {
					return true;
				}
			}
		}

		return false;
	}

	private void printStatistics() {
		int messagesRecieved = 0;
		int messagesProcessed = 0;
		int actorCount = 0;

		synchronized( theater.getRegistry() ) {
			Vector registry = new Vector(theater.getRegistry().values());

			for (int i = 0; i < registry.size(); i++) {
				Object current = registry.get(i);
				if (current instanceof RandomActorState) {
					messagesRecieved += ((RandomActorState)current).getRecieved();
					messagesProcessed += ((RandomActorState)current).getProcessed();
					actorCount++;
				}
			}
		}

		System.out.println(System.currentTimeMillis() + "\t" + messagesProcessed + "\t\t" + 
				   messagesRecieved + "\t\t" + actorCount);
	}

	public void run() {
		try {
			sleep(5000);
		} catch (InterruptedException ie) {
			System.err.println("Sleep interrupted!\n");
		}

		System.out.println("Time\t\tProcessed\tRecieved\tRegistrySize");

		//loop forever
		while (true) {
			if (!overloaded() && !waiting) {
				beginSteal();
			}

			printStatistics();
			reset();

			try {
				sleep(5000);
			} catch (InterruptedException ie) {
				System.err.println("Sleep interrupted!\n");
			}
		}
	}
}
